__all__ = (
    'PandakeeperError',
    'LoopedGraphError'
)


class PandakeeperError(Exception):
    """Base class for library-specific exceptions."""
    __slots__ = ()


class LoopedGraphError(PandakeeperError):
    """Throws when the logic of the algorithms can be violated by the presence of loops in graphs."""
    __slots__ = ()



from io import RawIOBase, BufferedIOBase, TextIOBase, TextIOWrapper
from mmap import mmap
from os import PathLike
from typing import Union, IO

__all__ = (
    'PD_READ_PICKLE_ANNOTATION',
)

PD_READ_PICKLE_ANNOTATION = Union[
    PathLike,
    str,
    IO[str],
    RawIOBase,
    BufferedIOBase,
    TextIOBase,
    TextIOWrapper,
    mmap
]



from itertools import chain, repeat
from types import MappingProxyType
from typing import Optional, Union, Dict, List, Tuple, Iterator, Iterable, Mapping

import pandas as pd
from pandera import DataFrameSchema
from typing_extensions import final
from varutils.typing import check_type_compatibility

from pandakeeper.node import Node
from pandakeeper.validators import AnyDataFrame

__all__ = (
    'NodeConnection',
    'DataProcessor'
)


class NodeConnection:
    """Class that encapsulates a connection to an input Node."""
    __slots__ = ('__node', '__input_validator')

    def __init__(self, node: Node, input_validator: DataFrameSchema = AnyDataFrame) -> None:
        """
        Class that encapsulates a connection to an input Node.

        Args:
            node:             input Node to connect to.
            input_validator:  DataFrameSchema that validates the data coming from the input Node.
        """
        check_type_compatibility(node, Node)
        check_type_compatibility(input_validator, DataFrameSchema)
        self.__node = node
        self.__input_validator = input_validator

    @final
    @property
    def node(self) -> Node:
        """Input Node."""
        return self.__node

    @final
    @property
    def input_validator(self) -> DataFrameSchema:
        """Input validator."""
        return self.__input_validator

    @final
    def extract_data(self) -> pd.DataFrame:
        """Extracts and validates data from the input Node."""
        data = self.__node.extract_data()
        return self.__input_validator.validate(data)


class DataProcessor(Node):
    """
    Abstract class that defines an interface common to all data processors,
    i.e. Nodes that can receive data from other Nodes through NodeConnection instances.
    """
    __slots__ = ('__positional_node_connections', '__named_node_connections')

    def __init__(self, output_validator: DataFrameSchema = AnyDataFrame):
        """
        Abstract class that defines an interface common to all data processors,
        i.e. Nodes that can receive data from other Nodes through NodeConnection instances.

        Args:
            output_validator: DataFrameSchema that validates the data coming from the 'extract_data' method.
        """
        super().__init__(output_validator)
        self.__positional_node_connections: List[NodeConnection] = []
        self.__named_node_connections: Dict[str, NodeConnection] = {}

    @final
    @property
    def positional_input_nodes(self) -> Tuple[NodeConnection, ...]:
        """Returns positional input NodeConnections."""
        return tuple(self.__positional_node_connections)

    @final
    @property
    def named_input_nodes(self) -> Mapping[str, NodeConnection]:
        """Returns named input NodeConnections."""
        return MappingProxyType(self.__named_node_connections)

    @final
    def __connect_input_node_body(self,
                                  node_connection: NodeConnection,
                                  keyword: Optional[str] = None) -> None:
        """
        Supplemental method used bt the 'connect_input_node' and the 'connect_input_nodes' methods.

        Args:
            node_connection:  Node to connect to or NodeConnection to add.
            keyword:          Optional name of the NodeConnection.
        """
        if keyword is None:
            self.__positional_node_connections.append(node_connection)
        else:
            named_node_connections = self.__named_node_connections
            if keyword in named_node_connections:
                raise KeyError(f"Duplicate name '{keyword}' for named NodeConnection")
            named_node_connections[keyword] = node_connection
        self._add_edge_to_connection_graph(node_connection.node)

    @staticmethod
    def __check_node_connection(node: Union[Node, NodeConnection],
                                keyword: Optional[str],
                                arg_position: int = 0) -> Tuple[NodeConnection, Optional[str]]:
        """
        Supplemental method used by the '__check_node_connections' and the 'connect_input_node' methods.

        Args:
            node:          Node to connect to or NodeConnection to add.
            keyword:       Optional name of the NodeConnection.
            arg_position:  Argument position. See the context of the 'connect_input_nodes' method.

        Returns:
            (NodeConnection, input keyword)
        """
        if isinstance(node, Node):
            return NodeConnection(node), keyword
        elif isinstance(node, NodeConnection):
            return node, keyword
        if keyword is None:
            err_msg = f"Positional argument in position {arg_position} has incompatible type: {type(node)}"
        else:
            err_msg = f"Argument '{keyword}' has incompatible type: {type(node)}"
        raise TypeError(err_msg)

    @staticmethod
    def __check_node_connections(nodes: Iterable[Tuple[Optional[str], Union[Node, NodeConnection]]]) -> Iterator[
        Tuple[NodeConnection, Optional[str]]
    ]:
        """
        Supplemental method used by the 'connect_input_nodes' method.

        Args:
            nodes:  Iterable of Nodes to connect to or NodeConnections to add.

        Returns:
            Iterator of pairs (NodeConnection, input keyword)
        """
        check_node_connection = DataProcessor.__check_node_connection
        for i, (keyword, node) in enumerate(nodes):
            yield check_node_connection(node, keyword, i)

    @final
    def connect_input_node(self,
                           node_connection: Union[Node, NodeConnection],
                           keyword: Optional[str] = None) -> None:
        """
        Connects input Node.

        Args:
            node_connection:  Node to connect to or NodeConnection to add.
            keyword:          Optional name of the NodeConnection.
        """
        args = DataProcessor.__check_node_connection(node_connection, keyword)
        self.drop_cache()
        self.__connect_input_node_body(*args)

    @final
    def connect_input_nodes(self,
                            *positional_nodes: Union[Node, NodeConnection],
                            **keyword_nodes: Union[Node, NodeConnection]) -> None:
        """
        Connects multiple input Nodes.

        Args:
            positional_nodes:  positional Nodes to connect or NodeConnections to add.
            keyword_nodes:     named Nodes to connect or NodeConnections to add.
        """
        if positional_nodes or keyword_nodes:
            nodes = tuple(
                DataProcessor.__check_node_connections(
                    chain(
                        zip(repeat(None), positional_nodes),
                        keyword_nodes.items()
                    )
                )
            )
            self.drop_cache()
            for args in nodes:
                self.__connect_input_node_body(*args)



from typing import Optional

from pandas import DataFrame
from typing_extensions import final

from pandakeeper.dataprocessor import DataProcessor

__all__ = (
    'DataCacher',
    'RuntimeCacher',
    'SingleInputRuntimeCacher',
)


class DataCacher(DataProcessor):
    """Abstract DataProcessor that caches input data."""
    __slots__ = ()

    @final
    @property
    def use_cached(self) -> bool:
        return True


class RuntimeCacher(DataCacher):
    """Abstract DataCacher for caching Node outputs to RAM."""
    __slots__ = ('__dataframe',)
    __dataframe: Optional[DataFrame]

    @final
    def _dump_to_cache(self, data: DataFrame) -> None:
        self.__dataframe = data

    @final
    def _clear_cache_storage(self) -> None:
        self.__dataframe = None

    @final
    def _load_cached(self) -> DataFrame:
        df = self.__dataframe
        if df is not None:
            return df
        raise ValueError("Cannot load non-cached data")


class SingleInputRuntimeCacher(RuntimeCacher):
    """RuntimeCacher for caching single input Node."""
    __slots__ = ()

    def _load_non_cached(self) -> DataFrame:
        pnc = self.positional_input_nodes
        nnc = self.named_input_nodes
        total_dfs = len(nnc) + len(pnc)
        if total_dfs != 1:
            raise ValueError(
                "Cannot be connected to more or less than one Node. "
                f"Actual number of connections: {total_dfs}"
            )
        try:
            node_connection = pnc[0]
        except IndexError:
            node_connection = next(iter(nnc.values()))
        return node_connection.extract_data()

    def transform_data(self, data: DataFrame) -> DataFrame:
        return data



from collections.abc import Mapping as _Mapping, Callable as _Callable
from contextlib import ExitStack
from types import MappingProxyType
from typing import Callable, Any, Tuple, Mapping

import pandas as pd
import pandera as pa
from typing_extensions import final
from varutils.plugs.constants import empty_mapping_proxy
from varutils.typing import check_type_compatibility

from pandakeeper.dataloader.core import StaticDataLoader

__all__ = ('SqlLoader',)


class SqlLoader(StaticDataLoader):
    """DataLoader that loads data using SQL-connections."""
    __slots__ = ('__context_creator', '__read_sql_fn')

    def __init__(
            self,
            context_creator: Callable[..., Any],
            sql_query: str,
            *,
            context_creator_args: Tuple[Any, ...] = (),
            context_creator_kwargs: Mapping[str, Any] = empty_mapping_proxy,
            read_sql_fn: Callable[..., pd.DataFrame] = pd.read_sql,
            read_sql_args: Tuple[Any, ...] = (),
            read_sql_kwargs: Mapping[str, Any] = empty_mapping_proxy,
            output_validator: pa.DataFrameSchema) -> None:
        """
        DataLoader that loads data using SQL-connections.

        Args:
            context_creator:         callable that should create a SQL context and return a DB-connection.
            sql_query:               SQL-query to run.
            context_creator_args:    positional arguments for 'context_creator'.
            context_creator_kwargs:  keyword arguments for 'context_creator'.
            read_sql_fn:             function that creates pandas.DataFrame from the result of SQL-query.
            read_sql_args:           positional arguments for 'read_sql_fn'.
            read_sql_kwargs:         keyword arguments for 'read_sql_fn'.
            output_validator:        output validator.
        """
        check_type_compatibility(context_creator, _Callable, 'Callable')  # type: ignore
        check_type_compatibility(sql_query, str)
        check_type_compatibility(context_creator_args, tuple)
        check_type_compatibility(context_creator_kwargs, _Mapping, "dict or another Mapping")
        check_type_compatibility(read_sql_fn, _Callable, 'Callable')  # type: ignore
        check_type_compatibility(read_sql_args, tuple)
        check_type_compatibility(read_sql_kwargs, _Mapping, "dict or another Mapping")
        super().__init__(
            self.__load_sql,
            sql_query,
            context_creator_args,
            context_creator_kwargs,
            read_sql_args,
            read_sql_kwargs
        )
        self.set_output_validator(output_validator)
        self.__read_sql_fn = read_sql_fn
        self.__context_creator = context_creator

    @final
    def __load_sql(
            self,
            sql_query: str,
            context_creator_args: Tuple[Any, ...],
            context_creator_kwargs: Mapping[str, Any],
            read_sql_args: Tuple[Any, ...],
            read_sql_kwargs: Mapping[str, Any]) -> pd.DataFrame:
        """
        Builds necessary contexts and returns the result of the SQL-query.

        Args:
            sql_query:               SQL-query to run.
            context_creator_args:    positional arguments for 'context_creator'.
            context_creator_kwargs:  keyword arguments for 'context_creator'.
            read_sql_args:           positional arguments for 'read_sql_fn'.
            read_sql_kwargs:         keyword arguments for 'read_sql_fn'.

        Returns:
            Resulting DataFrame.
        """
        with ExitStack() as exit_stack:
            conn = self.__context_creator(exit_stack, *context_creator_args, **context_creator_kwargs)
            return self.__read_sql_fn(sql_query, conn, *read_sql_args, **read_sql_kwargs)

    @final
    @property
    def _read_sql_fn(self) -> Callable[..., pd.DataFrame]:
        """Function that creates pandas.DataFrame from the result of SQL-query."""
        return self.__read_sql_fn

    @final
    @property
    def _context_creator(self) -> Callable[..., Any]:
        """Callable that create SQL context and return the DB-connection."""
        return self.__context_creator

    @final
    @property
    def _context_creator_args(self) -> Tuple[Any, ...]:
        """Positional arguments for 'sql_context_creator'."""
        return self._loader_args[1]

    @final
    @property
    def _context_creator_kwargs(self) -> Mapping[str, Any]:
        """Keyword arguments for 'sql_context_creator'."""
        return MappingProxyType(self._loader_args[2])

    @final
    @property
    def _read_sql_args(self) -> Tuple[Any, ...]:
        """Positional arguments for 'read_sql_fn'."""
        return self._loader_args[3]

    @final
    @property
    def _read_sql_kwargs(self) -> Mapping[str, Any]:
        """Keyword arguments for 'read_sql_fn'."""
        return MappingProxyType(self._loader_args[4])

    @final
    @property
    def sql_query(self) -> str:
        """SQL-query to run."""
        return self._loader_args[0]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pandakeeper',
 'pandakeeper.dataloader',
 'pandakeeper.dataloader.sql',
 'pandakeeper.dataprocessor']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1,<2',
 'pandera>=0.9',
 'typing-extensions>=4',
 'varname>=0.8,<0.9',
 'varutils>=0.0.8,<0.0.9']

setup_kwargs = {
    'name': 'pandakeeper',
    'version': '0.0.27',
    'description': 'Python library designed to simplify the validation of dataset manipulations.',
    'long_description': '# pandakeeper\nPython library designed to simplify the validation of dataset manipulations.\n',
    'author': 'Andrew Sonin',
    'author_email': 'sonin.cel@yandex.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/andrewsonin/pandakeeper',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from collections.abc import Callable as _Callable
from types import MappingProxyType
from typing import Any, Optional, Mapping, Callable, Tuple
from warnings import warn

import pandera as pa
from pandas import DataFrame, read_pickle, read_excel, read_csv
from typing_extensions import final
from varutils.plugs.functional import pass_through_one
from varutils.typing import check_type_compatibility

from pandakeeper.node import Node
from pandakeeper.typing import PD_READ_PICKLE_ANNOTATION
from pandakeeper.validators import AnyDataFrame

__all__ = (
    'DataLoader',
    'StaticDataLoader',
    'DataFrameAdapter',
    'PickleLoader',
    'ExcelLoader',
    'CsvLoader'
)


class DataLoader(Node):
    """
    Abstract class that defines an interface common to all data loaders,
    i.e. Nodes that can only generate data but not receive from other Nodes.
    """
    __slots__ = ('__loader', '__loader_args', '__loader_kwargs')

    def __init__(self,
                 loader: Callable[..., DataFrame],
                 *loader_args: Any,
                 **loader_kwargs: Any) -> None:
        """
        Abstract class that defines an interface common to all data loaders,
        i.e. Nodes that can only generate data but not receive from other Nodes.

        Args:
            loader:           loader function.
            *loader_args:     positional arguments passed to the loader function.
            **loader_kwargs:  keyword arguments passed to the loader function.
        """
        check_type_compatibility(loader, _Callable, 'Callable')  # type: ignore
        super().__init__(AnyDataFrame)
        self.__loader = loader
        self.__loader_args = loader_args
        self.__loader_kwargs = loader_kwargs

    @final
    def _load_default(self) -> DataFrame:
        """
        Returns the result of the loader function.

        Returns:
            Resulting DataFrame.
        """
        return self.__loader(*self.__loader_args, **self.__loader_kwargs)

    @final
    @property
    def _loader(self) -> Callable[..., DataFrame]:
        """Loader function."""
        return self.__loader

    @final
    @property
    def _loader_args(self) -> Tuple[Any, ...]:
        """Positional arguments of the loader function."""
        return self.__loader_args

    @final
    @property
    def _loader_kwargs(self) -> Mapping[str, Any]:
        """Keyword arguments of the loader function."""
        return MappingProxyType(self.__loader_kwargs)


class StaticDataLoader(DataLoader):
    """DataLoader base class for static data sources."""
    __slots__ = ()

    def _dump_to_cache(self, data: DataFrame) -> None:
        warn("'_dump_to_cache' does nothing for StaticDataLoader instances", RuntimeWarning)

    def _load_cached(self) -> DataFrame:
        warn(
            "'_load_cached' should not be called for StaticDataLoader instances. Switch to '_load_non_cached'",
            RuntimeWarning
        )
        return self._load_default()

    @final
    def _load_non_cached(self) -> DataFrame:
        return self._load_default()

    def _clear_cache_storage(self) -> None:
        warn("'_clear_cache_storage' does nothing for StaticDataLoader instances", RuntimeWarning)

    @property
    def use_cached(self) -> bool:
        return False

    @final
    def transform_data(self, data: DataFrame) -> DataFrame:
        return data


class DataFrameAdapter(StaticDataLoader):
    """DataLoader adapter for existing DataFrames."""
    __slots__ = ('__copy',)

    def __init__(self,
                 df: DataFrame,
                 *,
                 output_validator: pa.DataFrameSchema = AnyDataFrame,
                 copy: bool = False) -> None:
        """
        DataLoader adapter for existing DataFrames.

        Args:
            df:                input DataFrame.
            output_validator:  output validator.
            copy:              whether to copy input DataFrame.
        """
        check_type_compatibility(df, DataFrame)
        check_type_compatibility(copy, bool)
        if copy:
            df = df.copy()
        super().__init__(pass_through_one, df)
        self.set_output_validator(output_validator)
        self.__copy = copy

    @final
    @property
    def copy(self) -> bool:
        return self.__copy

    @final
    @property
    def dataframe(self) -> DataFrame:
        return self._loader_args[0]


class PickleLoader(StaticDataLoader):
    """DataLoader that loads pickled DataFrames."""
    __slots__ = ()

    def __init__(self,
                 filepath_or_buffer: PD_READ_PICKLE_ANNOTATION,
                 compression: Optional[str] = 'infer',
                 *,
                 output_validator: pa.DataFrameSchema = AnyDataFrame) -> None:
        """
        DataLoader that loads pickled DataFrames.

        Args:
            filepath_or_buffer:  filepath or buffer to read pickle from.
            compression:         input file compression.
            output_validator:    output validator.
        """
        super().__init__(read_pickle, filepath_or_buffer, compression)
        self.set_output_validator(output_validator)

    @final
    @property
    def filepath_or_buffer(self) -> PD_READ_PICKLE_ANNOTATION:
        return self._loader_args[0]

    @final
    @property
    def compression(self) -> Optional[str]:
        return self._loader_args[1]


class ExcelLoader(StaticDataLoader):
    """DataLoader that loads Excel files."""
    __slots__ = ()

    def __init__(self,
                 io,
                 *loader_args: Any,
                 output_validator: pa.DataFrameSchema = AnyDataFrame,
                 **loader_kwargs: Any) -> None:
        """
        DataLoader that loads Excel files.

        Args:
            io:                pandas.read_excel first argument.
            *loader_args:      pandas.read_excel positional arguments.
            output_validator:  output validator.
            **loader_kwargs:   pandas.read_excel keyword arguments.
        """
        super().__init__(read_excel, io, *loader_args, **loader_kwargs)
        self.set_output_validator(output_validator)

    @final
    @property
    def io(self):
        return self._loader_args[0]


class CsvLoader(StaticDataLoader):
    """DataLoader that loads csv-files."""
    __slots__ = ()

    def __init__(self,
                 filepath_or_buffer,
                 *loader_args: Any,
                 output_validator: pa.DataFrameSchema = AnyDataFrame,
                 **loader_kwargs: Any) -> None:
        """
        DataLoader that loads csv-files.

        Args:
            filepath_or_buffer:  pandas.read_csv first argument.
            *loader_args:        pandas.read_csv positional arguments.
            output_validator:    output validator.
            **loader_kwargs:     pandas.read_csv keyword arguments.
        """
        super().__init__(read_csv, filepath_or_buffer, *loader_args, **loader_kwargs)
        self.set_output_validator(output_validator)

    @final
    @property
    def filepath_or_buffer(self):
        return self._loader_args[0]



from abc import ABCMeta, abstractmethod
from collections import defaultdict
from typing import Dict, Set
from warnings import warn

import pandas as pd
from pandera import DataFrameSchema
from typing_extensions import final
from varutils.typing import check_type_compatibility

from pandakeeper.errors import LoopedGraphError

__all__ = ('Node',)


class Node(metaclass=ABCMeta):
    """Abstract class that defines an interface common to all data manipulators."""

    __slots__ = ('__gateway_id', '__already_cached', '__output_validator')
    __instance_counter = 0
    __parental_graph: Dict['Node', Set['Node']] = defaultdict(set)
    __children_graph: Dict['Node', Set['Node']] = defaultdict(set)

    def __init__(self, output_validator: DataFrameSchema) -> None:
        """
        Abstract class that defines an interface common to all data manipulators.

        Args:
            output_validator: DataFrameSchema that validates the data coming from the 'extract_data' method.
        """
        check_type_compatibility(output_validator, DataFrameSchema)
        self.__gateway_id = Node.__instance_counter
        Node.__instance_counter += 1
        self.__already_cached = False
        self.__output_validator = output_validator

    @final
    def __hash__(self) -> int:
        return self.__gateway_id

    @final
    @property
    def _output_validator(self) -> DataFrameSchema:
        """
        Returns output validator.

        Returns:
            pandera.DataFrameSchema that validates the result of the 'transform_data' method.
        """
        return self.__output_validator

    @final
    @property
    def _is_parental_graph_topo_sorted(self) -> bool:
        """
        Checks whether parental graph is topologically sorted.

        Returns:
            Result of checking.
        """
        connection_graph = Node.__parental_graph

        visited_nodes = {self}
        nodes_to_visit = connection_graph[self].copy()
        try:
            while True:
                cur_node = nodes_to_visit.pop()
                if cur_node in visited_nodes:
                    return False
                visited_nodes.add(cur_node)
                nodes_to_visit |= connection_graph[cur_node]
        except KeyError:
            return True

    @final
    def _add_edge_to_connection_graph(self, parent_node: 'Node') -> None:
        """
        Adds directed edge {Self <- Parent} to the connection graph.

        Args:
            parent_node:  parent Node to connect self Node to.
        """
        Node.__parental_graph[self].add(parent_node)
        Node.__children_graph[parent_node].add(self)

    @final
    def _remove_edge_from_connection_graph(self, parent_node: 'Node') -> None:
        """
        Removes directed edge {Self <- Parent} from the connection graph.

        Args:
            parent_node:  parent Node to untie self Node from.
        """
        Node.__parental_graph[self].remove(parent_node)
        Node.__children_graph[parent_node].remove(self)

    @final
    @property
    def already_cached(self) -> bool:
        """Checks whether self is already cached."""
        return self.__already_cached

    @final
    @property
    def gateway_id(self) -> int:
        """Returns unique ID of the Node instance."""
        return self.__gateway_id

    @final
    def __make_node_cached(self) -> None:
        """Supplemental method for the 'make_node_cached' method."""
        data = self._load_non_cached()
        data = self.transform_data(data)
        data = self.__output_validator.validate(data)
        self._dump_to_cache(data)
        self.__already_cached = True

    @final
    def make_node_cached(self) -> None:
        """Caches self and all parent Nodes with True use_cached property."""
        if self.__already_cached:
            return
        if not self._is_parental_graph_topo_sorted:
            node_id = self.__gateway_id
            raise LoopedGraphError(f"Parental graph of Node (ID={node_id}) has loops")
        for parent in Node.__parental_graph[self]:
            if parent.use_cached and not parent.__already_cached:
                parent.__make_node_cached()
        if not self.use_cached:
            node_id = self.__gateway_id
            warn(f"'make_node_cached' called for Node (ID={node_id}) with False 'use_cached' property", RuntimeWarning)
            return
        self.__make_node_cached()

    @final
    def drop_cache(self) -> None:
        """Drops the Node's cache, dropping it in all child Nodes as well."""
        if self.__already_cached:
            self._clear_cache_storage()
            self.__already_cached = False

        children_graph = Node.__children_graph
        visited_nodes = {self}
        nodes_to_visit = children_graph[self].copy()
        try:
            while True:
                cur_node = nodes_to_visit.pop()
                if cur_node in visited_nodes:
                    node_id = cur_node.__gateway_id
                    raise LoopedGraphError(f"Node with ID={node_id} is child of itself")
                if cur_node.__already_cached:
                    cur_node._clear_cache_storage()
                    cur_node.__already_cached = False
                visited_nodes.add(cur_node)
                nodes_to_visit |= children_graph[cur_node]
        except KeyError:
            pass

    @final
    def extract_data(self) -> pd.DataFrame:
        """
        Extracts data from the Node.

        Returns:
            Extracted DataFrame.
        """
        if self.__already_cached:
            data = self._load_cached()
            return self.__output_validator.validate(data)
        if not self._is_parental_graph_topo_sorted:
            node_id = self.__gateway_id
            raise LoopedGraphError(f"Parental graph of Node (ID={node_id}) has loops")
        data = self._load_non_cached()
        data = self.transform_data(data)
        data = self.__output_validator.validate(data)
        if self.use_cached:
            self._dump_to_cache(data)
            self.__already_cached = True
        return data

    @final
    def set_output_validator(self, output_validator: DataFrameSchema) -> 'Node':
        """
        Sets the output validator.

        Args:
            output_validator: output validator to set.
        Returns:
            Self instance.
        """
        check_type_compatibility(output_validator, DataFrameSchema)
        self.__output_validator = output_validator
        return self

    @abstractmethod
    def _dump_to_cache(self, data: pd.DataFrame) -> None:
        """
        Defines the last logical step of the 'extract_data' method. Dumps extracted data to cache.

        Args:
            data: DataFrame to dump.
        """

    @abstractmethod
    def _clear_cache_storage(self) -> None:
        """Clears cache storage. Must be the inverse of '_dump_to_cache'. Used by the 'drop_cache' method."""

    @abstractmethod
    def _load_cached(self) -> pd.DataFrame:
        """
        Loads data previously dumped by '_dump_to_cache' method.

        Returns:
            Loaded data.
        """

    @abstractmethod
    def _load_non_cached(self) -> pd.DataFrame:
        """
        Loads raw input data. Used by the 'extract_data' as the first logical step if not self.already_cached.

        Returns:
            Loaded data.
        """

    @property
    @abstractmethod
    def use_cached(self) -> bool:
        """Whether to cache the result of the 'extract_data' method or run it from scratch each time."""

    @abstractmethod
    def transform_data(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Applies transformations to the output of the '_load_non_cached' method. Used by the 'extract_data'
        as the last logical step before sending the data to the output validator for validation.

        Args:
            data: the output of the '_load_non_cached' method.
        Returns:
            Transformed output, ready to be validated by the output validator.
        """
